# Copyright © 2024 Yannick Gingras <ygingras@ygingras.net> and contributors

# This file is part of Minute Maze.

# Minute Maze is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Minute Maze is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Minute Maze.  If not, see <https://www.gnu.org/licenses/>.

## A thin layer of abstraction over TileMapLayer
class_name BoardLayer extends TileMapLayer
const TILE_SIDE = 32
const ATLAS_TILES = 0
const SCENE_TILES = 1

## {"name" -> "id"}
const TERRAIN_IDS = {"wall": 0, "floor":1}
const TERRAIN_CHARS = {"wall": "█", "floor": "."}

const ADJ_OFFSETS = [Vector2i(-1, -1), Vector2i(0, -1), Vector2i(1, -1), Vector2i(1, 0),
					Vector2i(1, 1), Vector2i(0, 1), Vector2i(-1, 1), Vector2i(-1, 0)]
const CROSS_OFFSETS = [Vector2i(0, -1), Vector2i(1, 0), Vector2i(0, 1), Vector2i(-1, 0)]

var rect : Rect2i
var alt_names := {}


static func cheby_dist(coord1:Vector2i, coord2:Vector2i) -> int:
	## Return the Chebyshev distance between two coords
	# slightly optimized expression for:
	#   max(abs(coord1.x - coord2.x), abs(coord1.y - coord2.y))
	var diff:Vector2i = (coord2 - coord1).abs()
	return diff[diff.max_axis_index()]

static func man_dist(coord1:Vector2i, coord2:Vector2i) -> int:
	## Return the Manhattan distance between two coords
	# slightly optimized expresson for:
	#   abs(coord1.x - coord2.x) + abs(coord1.y - coord2.y)
	var diff:Vector2i = (coord2 - coord1).abs()
	return diff.x + diff.y

static func supercell_str(coord:Vector2i):
	## Return a short hand notation of a supercell coord that is different from Vector2i.to_string()
	return "⟦%d¦%d⟧" % [coord.x, coord.y]

static func coord_to_pos(coord:Vector2i, centered:=true) -> Vector2:
	## Convert a tile coord into a pixel position.
	## `centered`: center of the tile if true, top-left otherwise
	var pos = Vector2(coord) * TILE_SIDE
	if centered:
		pos += Vector2.ONE * 0.5 * TILE_SIDE
	return pos

static func pos_to_coord(pos:Vector2, centered:=true) -> Vector2i:
	## Convert a pixel position into a tile coord.
	## `centered`: the center of the tile is considered if true, the top-left otherwise
	if centered:
		pos -= Vector2.ONE * 0.5 * TILE_SIDE
	return Vector2i(pos / TILE_SIDE)

static func _rect_coords(rect:Rect2i) -> Array:
	var coords = []
	for i in rect.size.x:
		for j in rect.size.y:
			coords.append(rect.position + Vector2i(i, j))
	return coords

static func _interpolate_path(path:Array[Vector2i]) -> Array[Vector2i]:
	## Return an array of coords that connect all the elements of `path`.
	## The Manhattan dist between successive coords will always be 1 (no diagonals).
	## Elements of path are included.
	var coords:Array[Vector2i] = [path[0]]
	var dir
	var diff
	for i in range(1, len(path)):
		var prev = coords[-1]
		if prev.x != path[i].x:
			diff = path[i].x - prev.x
			dir = sign(diff)
			for k in abs(diff) - 1:
				coords.append(coords[-1] + Vector2i.RIGHT * dir)
			if prev.y != path[i].y:
				coords.append(coords[-1] + Vector2i.RIGHT * dir)
		if prev.y != path[i].y:
			diff = path[i].y - prev.y
			dir = sign(diff)
			for k in abs(diff) - 1:
				coords.append(coords[-1] + Vector2i.DOWN * dir)
		coords.append(path[i])
	return coords

static func inner_rect(rect:Rect2i, margin=0):
	## Return a Rect2i that is one tile inside `rect`
	## margin: if >0, that many tiles separate the outer rect perimiter tiles from the inner rect.
	var delta = Vector2i.ONE * (1 + margin)
	assert(rect.size.x > delta.x*2 and rect.size.y > delta.y*2,
			"%s is too small to contain an inner rect" % rect)
	return Rect2i(rect.position+delta, rect.size-delta*2)

func add_tile(tile:Node, name:String):
	var source = tile_set.get_source(SCENE_TILES)
	var scene = PackedScene.new()
	scene.pack(tile)
	var id = source.create_scene_tile(scene)
	alt_names[name] = id

func paint_rect(rect:Rect2i, terrain:String):
	var coords = _rect_coords(rect)
	paint_coords(coords, terrain)

func paint_coords(coords:Array, terrain:String):
	if terrain in TERRAIN_IDS:
		var tid = TERRAIN_IDS[terrain]
		set_cells_terrain_connect(coords, ATLAS_TILES, tid, false)
	elif terrain in alt_names:
		for coord in coords:
			set_cell(coord, SCENE_TILES, Vector2i.ZERO, alt_names[terrain])
	else:
		assert(false, "can't find terrain '%s'" % terrain)

func paint_coord(coord:Vector2i, terrain:String):
	paint_coords([coord], terrain)

func paint_path(path:Array, terrain:String):
	# could also use set_cells_terrain_path(), but it's unclear if there if a benefit.
	var coords = _interpolate_path(path)
	paint_coords(coords, terrain)

func get_cell_terrain(coord):
	var data = get_cell_tile_data(coord) as TileData
	if data == null:
		# FIXME: could be a scene-tile, worth a lookup in alt_names
		return null
	else:
		return tile_set.get_terrain_name(data.terrain_set, data.terrain)

func is_walkable(coord:Vector2i):
	return get_cell_terrain(coord) == "floor"

func ddump():
	## Print a representation of this board to the console.
	var size = get_used_rect().size
	for j in size.y:
		var chars = []
		for i in size.x:
			var terrain = get_cell_terrain(Vector2i(i, j))
			chars.append(_terrain_char(terrain))
		print("".join(chars))

func _terrain_char(terrain) -> String:
	if terrain == null:
		return " "
	elif TERRAIN_CHARS.has(terrain):
		return TERRAIN_CHARS[terrain]
	else:
		return terrain[0]


func dijkstra(start:Vector2i, dest:=Consts.COORD_INVALID, visit_func=null):
#func dist_metrics(start_:Vector2i=Consts.COORD_INVALID, dest_:Vector2i=Consts.COORD_INVALID,
					#free_dest_:=false, max_dist:int=-1, max_steps:int=-1,
					#valid_pred=null, index=null):
	## Return distance metrics or all positions accessible from start.
	## visit_func: if provided, will be called every time we visit a cell.
	var rect = get_used_rect()
	var metrics = DistMetrics2i.new(rect.size, start)
	var queue = DistQueue2i.new()
	var max_dist = null
	var dist := 0
	var done = {}

	var known_good = {}  # used a a set

	queue.enqueue(start, dist)
	var current:Vector2i
	var entry: DistQueue2i.Entry
	var nb_steps := 0
	while not queue.is_empty():
		nb_steps += 1
		#if max_steps > 0 and nb_steps >= max_steps:
			#break
		entry = queue.dequeue()
		dist = entry.dist
		current = entry.coord
		if visit_func:
			visit_func.call(current)
		if current == dest:
			break  # Done!
		elif dest != Consts.COORD_INVALID and man_dist(current, dest) == 1:
			# got next to dest, no need to look at adjacents()
			metrics.setv(dest, dist+1)
			metrics.add_edge(current, dest)
			break
		if done.has(current) or dist == max_dist:
			continue  # this position is finalized already

		# This inline version of `adjacent_walkable(current, ctx.pred)`
		# is about twice as fast. It depends on ctx.pred() to be sensible.
		var adjs:Array[Vector2i] = []
		for offset in CROSS_OFFSETS:
			var adj = current + offset
			if not rect.has_point(adj) or done.has(adj):
				continue
			elif known_good.has(adj):
				adjs.append(adj)
			elif is_walkable(adj):
				known_good[adj] = true
				adjs.append(adj)

		for pos in adjs:
			if done.has(pos):
				continue
			var pre_dist = metrics.getv(pos)
			if pre_dist == -1 or pre_dist > dist+1:
				metrics.setv(pos, dist+1)
				metrics.add_edge(current, pos)
			queue.enqueue(pos, dist+1)
		done[current] = true
	return metrics

func astar(start:Vector2i, dest:Vector2i, visit_func=null):
	## Return sparse BoardMetrics using the A* algorithm.
	var rect = get_used_rect()
	var metrics = DistMetrics2i.new(rect.size, start)
	var queue = DistQueue2i.new()

	# The dist used in the queue is dist is h(n), then estimate from here to the destination.
	# Ties are broken with the [g(n)] pair: the real dist from start packed in a list because that's how the
	# custom fast queue expects it.
	var estimate:int = man_dist(start, dest)
	var tie_breaker:Array[int] = [0]
	var done = {}

	queue.enqueue(start, estimate, tie_breaker)
	var current:Vector2i
	var entry: DistQueue2i.Entry
	while not queue.is_empty():
		entry = queue.dequeue()
		tie_breaker = entry.tie_breaker
		current = entry.coord
		if visit_func:
			visit_func.call(current)
		if current == dest:
			break  # Done!
		elif man_dist(current, dest) == 1:
			# got next to dest, no need to look at adjacents()
			metrics.setv(dest, tie_breaker[0]+1)
			metrics.add_edge(current, dest)
			break
		if done.has(current):
			continue  # this position is finalized already

		for offset in CROSS_OFFSETS:
			var adj = current + offset
			if not rect.has_point(adj) or done.has(adj) or not is_walkable(adj):
				continue
			var pre_dist = metrics.getv(adj)
			if pre_dist == -1 or pre_dist > tie_breaker[0]+1:
				metrics.setv(adj, tie_breaker[0]+1)
				metrics.add_edge(current, adj)
			estimate = man_dist(adj, dest) + tie_breaker[0]
			queue.enqueue(adj, estimate, [tie_breaker[0]+1])
		done[current] = true
	return metrics
