# Copyright © 2024 – 2025 Yannick Gingras <ygingras@ygingras.net> and contributors

# This file is part of Minute Maze.

# Minute Maze is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Minute Maze is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Minute Maze.  If not, see <https://www.gnu.org/licenses/>.

class_name Main extends Node2D

const PREP_STEPS = ["3", "2", "1", "Go!"]
const SHOW_PATH = false

var size: Vector2i
@onready var tracer = %Character
var drag_origin = null
var drag_delta = null
var prep_step := 0
var music_pos : float  # only used when pausing the game
var prev_coord : Vector2i
var dist_from_start : DistMetrics2i
var dist_from_goal : DistMetrics2i
var has_strayed := false
var buttons_delta := Vector2.ZERO

func _ready():
	gen_scene_tiles()
	%HUD/VersionLabel.text = "v" + Consts.VERSION

	var board_size = next_size()
	var view_size = get_viewport_rect().size

	# When button dpad is active, see how much smaller the board height has to be
	var settings = load_settings()
	apply_settings(settings)
	if settings.input_style == Consts.InputStyles.DPAD:
		var dpad_ratio = %DPad.size.y / view_size.y
		board_size.y = ceil(board_size.y * (1.0 - dpad_ratio))
	
	var scale = 1.0 * view_size.x / board_size.x / BoardLayer.TILE_SIDE
	for layer in [%Board, %Highlights]:
		layer.transform = layer.transform.scaled(Vector2.ONE * scale)
	var rect = Rect2i(Vector2i.ZERO, board_size)
	%Board.clear()
	%Board.paint_rect(rect, "wall")

	# fill with the maze pattern
	var biases = {"branching"=0.5, "reconnect"=0.1, "twistiness"=0.7}
	var mazer = Mazes.GrowingTree.new(%Board, biases, BoardLayer.inner_rect(rect), false)
	var seed = mazer.rand_supercell()
	mazer.fill(seed)
	%Board.ddump()

	# place player and goal as far apart from each other as possible
	var metrics = %Board.dijkstra(seed)
	var goal_coord = Vector2(metrics.furthest_coord)
	dist_from_goal = %Board.dijkstra(goal_coord)
	var start_coord = Vector2(dist_from_goal.furthest_coord)
	tracer.position = BoardLayer.coord_to_pos(start_coord)
	dist_from_start = %Board.dijkstra(start_coord)

	prev_coord = start_coord
	%Goal.position = BoardLayer.coord_to_pos(goal_coord)

	# timers and annoucements management
	if Globals.secs_left:
		%RaceTimer.wait_time = Globals.secs_left
	flash_cur_prep_step()
	%PrepTimer.start()


func _unhandled_input(event: InputEvent) -> void:
	if event.is_action_pressed("Test"):
		test()
		get_viewport().set_input_as_handled()
	elif not is_game_active() or Globals.settings.input_style == Consts.InputStyles.DPAD:
		get_viewport().set_input_as_handled()
	elif event is InputEventMouseButton and event.is_pressed() and event.button_index == MOUSE_BUTTON_LEFT:
		%VJoyAnchor.position = event.position
		%VJoyAnchor.show()
		%VJoyStick.position = event.position
		%VJoyStick.show()
		drag_origin = event.position
		drag_delta = null
		get_viewport().set_input_as_handled()
	elif event is InputEventMouseButton and not event.is_pressed() and event.button_index == MOUSE_BUTTON_LEFT:
		%VJoyAnchor.hide()
		%VJoyStick.hide()
		drag_origin = null
		drag_delta = null
		get_viewport().set_input_as_handled()
	elif event is InputEventMouseMotion and event.button_mask & MOUSE_BUTTON_MASK_LEFT and drag_origin:
		drag_delta = event.position - drag_origin
		%VJoyStick.position = event.position
		get_viewport().set_input_as_handled()

func _process(_delta: float):
	if %RaceTimer.time_left:
		%HUD/TimeLeftLabel.text = Utils.mins_str(%RaceTimer.time_left)

func _physics_process(_delta: float) -> void:
	if tracer != null and is_game_active():
		var pace
		if drag_delta:
			pace = drag_delta * drag_delta.length() ** Consts.VJOY_ACCEL
			tracer.set_pace(pace)
		else:
			var input_dir = Input.get_vector("ui_left", "ui_right", "ui_up", "ui_down")
			pace = (buttons_delta + input_dir) * Globals.settings.speed
			tracer.set_pace(pace)
		tracer.move_and_slide()
	
	# FIXME: gate by non-zero pace
	var cur_coord = BoardLayer.pos_to_coord(tracer.position, false)
	if cur_coord != prev_coord and not is_best_move(prev_coord, cur_coord):
		if not has_strayed:
			%StraySound.play()
			has_strayed = true
		print("bad move from %s to %s" % [cur_coord, prev_coord])
	prev_coord = cur_coord

	if SHOW_PATH:
		%Highlights.paint_coord(cur_coord, "dyn")

func load_settings() -> Settings:
	var settings:Settings
	if FileAccess.file_exists(Consts.SETTINGS_PATH):
		settings = ResourceLoader.load(Consts.SETTINGS_PATH, "", ResourceLoader.CACHE_MODE_IGNORE)
	else:
		settings = Settings.new()
		settings.save()

	Globals.settings = settings
	return settings

func apply_settings(settings:Settings):
	if settings.input_style == Consts.InputStyles.DPAD:
		%DPad.show()
	else:
		%DPad.hide()	

func is_game_active() -> bool:
	return not (%RaceTimer.is_stopped() or %RaceTimer.paused)

func start_race():
	%StartTone.play()
	%RaceTimer.start()
	await %StartTone.finished
	%RaceMusic.play()

func pause():
	## pause the game, regartedless of whether the race has started or not
	music_pos = %RaceMusic.get_playback_position()
	%PrepTimer.paused = true
	%RaceTimer.paused = true
	%RaceMusic.stop()

func resume():
	## resume the game, regartedless of whether the race has started or not
	%IdleMusic.stop()
	%PrepTimer.paused = false
	%RaceTimer.paused = false
	if is_game_active():
		%RaceMusic.play(music_pos)

func flash_cur_prep_step():
	if prep_step < len(PREP_STEPS) - 1:
		%PrepTone.play()

	%HUD/PrepLabel.modulate.a = 0
	%HUD/PrepLabel.text = PREP_STEPS[prep_step]
	var anim = create_tween()
	anim.tween_property(%HUD/PrepLabel, "modulate:a", 1, 0.1)
	anim.set_ease(Tween.EASE_IN).set_trans(Tween.TRANS_CUBIC)
	anim.tween_property(%HUD/PrepLabel, "modulate:a", 0, 0.9)

func is_best_move(from_coord, to_coord):
	## Return whether a move is optimal.
	## Because the mazes are not all perfect mazes, there can be more than one optimal moves from a given coord.
	return (dist_from_start.getv(from_coord) == dist_from_start.getv(to_coord) - 1
			and dist_from_goal.getv(from_coord) == dist_from_goal.getv(to_coord) + 1)

func _on_goal_body_entered(_body: Node2D) -> void:
	Globals.nb_solved += 1
	Globals.secs_left = %RaceTimer.time_left
	Globals.has_strayed = has_strayed
	get_tree().change_scene_to_file("res://src/ui/victory_screen.tscn")

func next_size() -> Vector2i:
	## Return the size of the next board to be generated in tiles
	var w:int = Consts.START_WIDTH + Globals.nb_solved * 2
	var h:int = w / get_viewport_rect().size.aspect()
	if not h % 2:
		h += 1
	return Vector2i(w, h)

func _on_prep_timer_timeout() -> void:
	prep_step += 1
	if prep_step < len(PREP_STEPS):
		flash_cur_prep_step()
		%PrepTimer.start()
	if prep_step == len(PREP_STEPS) - 1:
		# last prep message is currently flashing
		start_race()

func _on_race_timer_timeout() -> void:
	get_tree().change_scene_to_file("res://src/ui/game_over_screen.tscn")

func _on_help_button_pressed() -> void:
	pause()
	%HelpScreen.show()
	%IdleMusic.play()
	await %HelpScreen.done
	apply_settings(Globals.settings)
	resume()

func _on_dir_button_down(dir:Vector2i):
	buttons_delta += Vector2(dir)

func _on_dir_button_up(dir:Vector2i):
	buttons_delta -= Vector2(dir)

func gen_scene_tiles():
	## dynamically define some highlight types
	var tile = load("res://src/tile_highlight.tscn").instantiate()
	tile.modulate = "#00ff0080"
	%Highlights.add_tile(tile, "seen")
	tile.modulate = "#0000ff80"
	%Highlights.add_tile(tile, "unseen")

	tile = load("res://src/dynamic_highlight.tscn").instantiate()
	%Highlights.add_tile(tile, "dyn")

	tile = load("res://src/dynamic_highlight.tscn").instantiate()
	tile.on_screen_secs = 1.0
	tile.modulate = "#0090B4"
	%Highlights.add_tile(tile, "flash")

	tile = load("res://src/dynamic_highlight.tscn").instantiate()
	tile.modulate = "#ffaaaa"
	tile.alpha = 0.4
	%Highlights.add_tile(tile, "path")


func anim_path_finding(metrics_func:Callable):
	## Show the internal order of cell processing for path finding with Dijkstra,
	## then animate the discovered path.
	const pre_path_secs = 0.5
	%Highlights.clear()
	var visited_cells = []
	var visit = func(coord):
		visited_cells.append(coord)
	var here = %Board.pos_to_coord(tracer.position, false)
	var there = %Board.pos_to_coord(%Goal.position, false)

	var metrics = metrics_func.call(here, there, visit)
	var path = metrics.path(there)

	# The timers are adding the highlighs because if we add them directly after awaiting by delay_step, we end up
	# only adding one per frame, probably because that's the max resolution for a conroutine context change.
	var delay_step = (Consts.PATH_ANIM_SECS - pre_path_secs) * 0.95 / len(visited_cells)
	var delay = 0.0

	# discovery
	for coord in visited_cells:
		var timer = get_tree().create_timer(delay)
		timer.timeout.connect(%Highlights.paint_coord.bind(coord, "flash"))
		delay += delay_step

	# path
	delay += pre_path_secs
	path.reverse()
	delay_step = (Consts.PATH_ANIM_SECS - pre_path_secs) * 0.05 / len(path)
	for coord in path:
		var timer = get_tree().create_timer(delay)
		timer.timeout.connect(%Highlights.paint_coord.bind(coord, "path"))
		delay += delay_step

	# let the caller know that anims are still in progress
	await get_tree().create_timer(delay).timeout

func test():
	print("Testing: 1, 2...    1, 2...")
	await anim_path_finding(%Board.dijkstra)
	await get_tree().create_timer(1.0).timeout
	await anim_path_finding(%Board.astar)
	await get_tree().create_timer(5.0).timeout
	%Highlights.clear()
